@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2022	PgPool Global Development Group
+ * Copyright (c) 2003-2024	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -44,6 +44,8 @@ static POOL_STATUS handle_mismatch_tuples(POOL_CONNECTION * frontend, POOL_CONNE
 static int	forward_command_complete(POOL_CONNECTION * frontend, char *packet, int packetlen);
 static int	forward_empty_query(POOL_CONNECTION * frontend, char *packet, int packetlen);
 static int	forward_packet_to_frontend(POOL_CONNECTION * frontend, char kind, char *packet, int packetlen);
+static void process_clear_cache(POOL_CONNECTION_POOL * backend);
+static void clear_query_cache(void);
 
 POOL_STATUS
 CommandComplete(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend, bool command_complete)
@@ -140,6 +142,36 @@ CommandComplete(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend, bool
 		memcpy(p1, p, len);
 	}
 
+	/*
+	 * Check if a transaction start command was issued by a multi statement
+	 * query by looking at the command tag. We cannot look into the parse tree
+	 * because our parse tree is only the first query in the multi statement.
+	 */
+	if (SL_MODE && session_context->query_context &&
+		session_context->query_context->is_multi_statement)
+	{
+		if (!strcmp(p1, "BEGIN"))
+		{
+			/*
+			 * If the query was a transaction starting command, remember it
+			 * until it gets committed or roll backed.
+			 */
+			elog(DEBUG1, "Call set_tx_started_by_multi_statement_query() in CommandComplete");
+			set_tx_started_by_multi_statement_query();
+		}
+		else if (!strcmp(p1, "COMMIT") || !strcmp(p1, "ROLLBACK"))
+		{
+			/*
+			 * It is possible that the multi statement query included both
+			 * BEGIN and COMMIT/ROLLBACK, or just COMMIT/ROLLBACK command.  In
+			 * this case we forget that a transaction was started by multi
+			 * statement query.
+			 */
+			elog(DEBUG1, "Call unset_tx_started_by_multi_statement_query() in CommandComplete");
+			unset_tx_started_by_multi_statement_query();
+		}
+	}
+
 	/*
 	 * If operated in streaming replication mode and extended query mode, just
 	 * forward the packet to frontend and we are done. Otherwise, we need to
@@ -334,11 +366,35 @@ handle_query_context(POOL_CONNECTION_POOL * backend)
 		{
 			/* Commit ongoing CREATE/DROP temp table status */
 			pool_temp_tables_commit_pending();			
+
+			/* Forget a transaction was started by multi statement query */
+			unset_tx_started_by_multi_statement_query();
+
+			/*
+			 * Query cache enabled and REVOKE was issued in this transaction?
+			 */
+			if (pool_config->memory_cache_enabled &&
+				query_cache_disabled_tx())
+			{
+				/*
+				 * Clear all the query cache.
+				 */
+				clear_query_cache();
+
+				/* Unset query cache disabled in this transaction */
+				unset_query_cache_disabled_tx();
+			}
 		}
 		else if (stmt->kind == TRANS_STMT_ROLLBACK)
 		{
 			/* Remove ongoing CREATE/DROP temp table status */
 			pool_temp_tables_remove_pending();
+
+			/* Forget a transaction was started by multi statement query */
+			unset_tx_started_by_multi_statement_query();
+
+			/* Unset query cache disabled in this transaction */
+			unset_query_cache_disabled_tx();
 		}
 	}
 	else if (IsA(node, CreateStmt))
@@ -397,6 +453,34 @@ handle_query_context(POOL_CONNECTION_POOL * backend)
 			}
 		}
 	}
+	else if (IsA(node, VariableSetStmt))
+	{
+		VariableSetStmt *stmt = (VariableSetStmt *) node;
+
+		/* SET ROLE or SESSION AUTHORIZATION command? */
+		if (stmt->kind == VAR_SET_VALUE &&
+			(!strcmp(stmt->name, "role") ||
+			 !strcmp(stmt->name, "session_authorization")))
+			/* disable query cache in this session */
+			set_query_cache_disabled();
+	}
+	else if (IsA(node, GrantStmt))
+	{
+		GrantStmt *stmt = (GrantStmt *) node;
+
+		/* REVOKE? */
+		if (stmt->is_grant)
+			return;
+
+		/* Clear query cache */
+		process_clear_cache(backend);
+	}
+	else if (IsA(node, AlterTableStmt) || IsA(node, AlterDatabaseStmt) ||
+			 IsA(node, AlterDatabaseSetStmt) || IsA(node, AlterRoleStmt))
+	{
+		/* Clear query cache */
+		process_clear_cache(backend);
+	}
 }
 
 /*
@@ -571,3 +655,103 @@ forward_packet_to_frontend(POOL_CONNECTION * frontend, char kind, char *packet,
 
 	return 0;
 }
+
+/*
+ * Process statements that need clearing query cache
+ */
+static void
+process_clear_cache(POOL_CONNECTION_POOL * backend)
+{
+	/* Query cache enabled? */
+	if (!pool_config->memory_cache_enabled)
+		return;
+
+	/* Streaming replication mode? */
+	if (SL_MODE)
+	{
+		/*
+		 * Are we inside a transaction?
+		 */
+		if (TSTATE(backend, MAIN_NODE_ID ) == 'T')
+		{
+			/*
+			 * Disable query cache in this transaction.
+			 * All query cache will be cleared at commit.
+			 */
+			set_query_cache_disabled_tx();
+		}
+		else if (TSTATE(backend, MAIN_NODE_ID ) == 'I')	/* outside transaction */
+		{
+			/*
+			 * Clear all the query cache.
+			 */
+			clear_query_cache();
+		}
+	}
+	else
+	{
+		/*
+		 * Are we inside a transaction?
+		 */
+		if (TSTATE(backend, MAIN_NODE_ID ) == 'T')
+		{
+			/* Inside user started transaction? */
+			if (!INTERNAL_TRANSACTION_STARTED(backend, MAIN_NODE_ID))
+			{
+				/*
+				 * Disable query cache in this transaction.
+				 * All query cache will be cleared at commit.
+				 */
+				set_query_cache_disabled_tx();
+			}
+			else
+			{
+				/*
+				 * Clear all the query cache.
+				 */
+				clear_query_cache();
+			}
+		}
+		else if (TSTATE(backend, MAIN_NODE_ID ) == 'I')	/* outside transaction */
+		{
+			/*
+			 * Clear all the query cache.
+			 */
+			clear_query_cache();
+		}
+	}
+}
+
+/*
+ * Clear query cache on shmem or memcached
+ */
+static
+void clear_query_cache(void)
+{
+	/*
+	 * Clear all the shared memory cache and oid maps.
+	 */
+	if (pool_is_shmem_cache())
+	{
+		pool_clear_memory_cache();
+		ereport(LOG,
+				(errmsg("all query cache in shared memory deleted")));
+	}
+	else
+#ifdef USE_MEMCACHED
+	{
+		/*
+		 * Clear all the memcached cache and oid maps.
+		 */
+		delete_all_cache_on_memcached();
+		pool_discard_oid_maps();
+		ereport(LOG,
+				(errmsg("all query cache in memcached deleted")));
+	}
+#else
+	{
+		ereport(WARNING,
+				(errmsg("failed to clear cache on memcached, memcached support is not enabled")));
+	}
+#endif
+}
